package pl.demo.form.product_details;

import javax.inject.Inject;

import com.reforms21.api.*;
import com.reforms21.api.common.*;
import com.reforms21.api.dbms.*;
import com.reforms21.api.annotation.*;
import com.reforms21.api.annotation.Package;
import com.reforms21.api.annotation.key.*;
import com.reforms21.api.exception.*;
import com.reforms21.api.lib.*;

import com.reforms21.common.*;
import com.reforms21.common.exception.*;

import static com.reforms21.api.Forms.*;
import static com.reforms21.api.Reforms.*;
import static com.reforms21.api.Standard.*;

@FormController("PRODUCT_DETAILS")
public class ProductDetailsForm extends pl.demo.form.BaseForm {

	@Inject OrderStatus order_status;
	
	@Inject Frm frm;
	
	@Inject	Utils utils;
	
	@Inject	AppRunner app_runner;
	
	@Inject	FormOperations form_operations;
	

	@OnClearDetails
	void form_ON_CLEAR_DETAILS() {
		//
		// Begin default relation program section
		//
		{
			clear_all_master_details();
		}
		//
		// End default relation program section
		//
	}

	@WhenNewFormInstance
	void form_WHEN_NEW_FORM_INSTANCE() {
		frm.new_form_instance();
	}

	@WhenWindowClosed
	void form_WHEN_WINDOW_CLOSED() {
		exitForm();
	}

	@WhenMenuSelected("DETAILS_POPUP_MENU.COPY")
	void item_DETAILS_POPUP_MENU_COPY_WHEN_MENU_SELECTED() {
		copyRegion();
	}

	@WhenMenuSelected("DETAILS_POPUP_MENU.PASTE")
	void item_DETAILS_POPUP_MENU_PASTE_WHEN_MENU_SELECTED() {
		copyRegion();
	}

	@WhenMenuSelected("DETAILS_POPUP_MENU.SAVE")
	void item_DETAILS_POPUP_MENU_SAVE_WHEN_MENU_SELECTED() {
		form_operations.commit_current_form();
		form_operations.exit_current_form();
	}

	@WhenMenuSelected("DETAILS_POPUP_MENU.EXIT")
	void item_DETAILS_POPUP_MENU_EXIT_WHEN_MENU_SELECTED() {
		form_operations.exit_current_form();
	}

	@OnCheckDeleteMaster("PRODUCT_INFORMATION")
	void block_PRODUCT_INFORMATION_ON_CHECK_DELETE_MASTER() {
		//
		// Begin default relation declare section
		//
		{
			SqlVarchar dummy_define;
			//
			// Begin ORDER_ITEMS detail declare section
			//
			Cursor<?> order_items_cur = cursorFor(
				"SELECT 1 FROM ORDER_ITEMS O \n" + 
				"WHERE O.PRODUCT_ID = :PRODUCT_INFORMATION.PRODUCT_ID"
			);
			
			//
			// End ORDER_ITEMS detail declare section
			//
			//
			// Begin INVENTORIES detail declare section
			//
			Cursor<?> inventories_cur = cursorFor(
				"SELECT 1 FROM INVENTORIES I \n" + 
				"WHERE I.PRODUCT_ID = :PRODUCT_INFORMATION.PRODUCT_ID"
			);
			//
			// End INVENTORIES detail declare section
			//
			//
			// End default relation declare section
			//
			//
			// Begin default relation program section
			//
			try {
				//
				// Begin ORDER_ITEMS detail program section
				//
				open(order_items_cur);
				fetchInto(order_items_cur, dummy_define);
				if (order_items_cur.found()) {
					message("Cannot delete master record when matching detail records exist.");
					close(order_items_cur);
					throw new FormTriggerFailureException();
				}
				close(order_items_cur);
				//
				// End ORDER_ITEMS detail program section
				//
				//
				// Begin INVENTORIES detail program section
				//
				open(inventories_cur);
				fetchInto(inventories_cur, dummy_define);
				if (inventories_cur.found()) {
					message("Cannot delete master record when matching detail records exist.");
					close(inventories_cur);
					throw new FormTriggerFailureException();
				}
				close(inventories_cur);
				//
				// End INVENTORIES detail program section
				//
			} finally {
				if (order_items_cur.isOpen()) {
					close(order_items_cur);
				}
				if (inventories_cur.isOpen()) {
					close(inventories_cur);
				}
			}
		}
		//
		// End default relation program section
		//

	}

	@OnPopulateDetails("PRODUCT_INFORMATION")
	void block_PRODUCT_INFORMATION_ON_POPULATE_DETAILS() {
		//
		// Begin default relation declare section
		//
		{
			SqlVarchar recstat = nameIn("System.record_status");
			SqlVarchar startitm = nameIn("System.cursor_item");
			Relation rel_id = Relation.NULL();
			//
			// End default relation declare section
			//
			//
			// Begin default relation program section
			//

			if (recstat == "NEW" || recstat == "INSERT") {
				return;
			}
			//
			// Begin ORDER_ITEMS detail program section
			//
			if (isNotNull(numberIn("PRODUCT_INFORMATION.PRODUCT_ID"))) {
				rel_id = findRelation("PRODUCT_INFORMATION.PRODUCT_INFORMA_ORDER_ITEMS");
				query_master_details(rel_id, "ORDER_ITEMS");
			}
			//
			// End ORDER_ITEMS detail program section
			//
			//
			// Begin INVENTORIES detail program section
			//
			if (isNotNull(numberIn("PRODUCT_INFORMATION.PRODUCT_ID"))) {
				rel_id = findRelation("PRODUCT_INFORMATION.PRODUCT_INFORMA_INVENTORIES");
				query_master_details(rel_id, "INVENTORIES");
			}
			//
			// End INVENTORIES detail program section
			//

			if (nameIn("System.cursor_item") != startitm) {
				goItem(startitm);
				check_package_failure();
			}
		}
		//
		// End default relation program section
		//

	}

	@PostQuery("ORDER_ITEMS")
	void block_ORDER_ITEMS_POST_QUERY() {
		frm.populate_orders_data();
	}

	@WhenValidateItem("ORDER_ITEMS.ORDER_DATE")
	void item_ORDER_ITEMS_ORDER_DATE_WHEN_VALIDATE_ITEM() {
		// override to skip ifeld validation --
		/* null; */
	}

	@PostQuery("INVENTORIES")
	void block_INVENTORIES_POST_QUERY() {
		frm.populate_warehouses_data();
	}

	@WhenButtonPressed("NAVIGATION.BACK_BUTTON")
	void item_NAVIGATION_BACK_BUTTON_WHEN_BUTTON_PRESSED() {
		{
			form_operations.exit_current_form();
		}
	}

	@WhenButtonPressed("NAVIGATION.SAVE_BUTTON")
	void item_NAVIGATION_SAVE_BUTTON_WHEN_BUTTON_PRESSED() {
		{
			form_operations.commit_current_form();
			form_operations.exit_current_form();
		}
	}

	@Package
	class OrderStatusBody extends OrderStatus {

		SqlNumber get_order_status(SqlVarchar p_order_status_name) {
			SqlNumber v_order_status;

			try {
				sqlSelectInto(
					v_order_status,
					"SELECT order_status FROM order_statuses_view WHERE order_status_name = :1",
					p_order_status_name
				);
				
				return v_order_status;
			} catch (NoDataFoundException e1) {
				return -1;
			}
		}

		SqlVarchar get_order_status_name(SqlNumber p_order_status) {
			SqlVarchar v_order_status_name;

			try {
				sqlSelectInto(
					v_order_status_name,
					"SELECT order_status_name FROM order_statuses_view WHERE order_status = :1",
					p_order_status
				);
				
				return v_order_status_name;
			} catch (NoDataFoundException e1) {
				return "UNKNOWN";
			}
		}

		SqlVarchar get_va_name(SqlNumber p_order_status_code) {
			SqlVarchar v_va_name;

			v_va_name = get_order_status_name(p_order_status_code);
			sqlSelectInto(v_va_name, "SELECT 'STATUS_' || REPLACE(:1, ' ', '_') || '_VA' FROM dual", v_va_name);
			return v_va_name;
		}

		void colour_row_by_status(SqlVarchar p_order_status_text_item_name, SqlNumber p_order_status_code) {
			setItemInstanceProperty(p_order_status_text_item_name, CURRENT_RECORD, VISUAL_ATTRIBUTE, get_va_name(p_order_status_code));
		}
	}

	abstract class OrderStatus {

		abstract SqlNumber get_order_status(SqlVarchar p_order_status_name);

		abstract SqlVarchar get_order_status_name(SqlNumber p_order_status);

		abstract SqlVarchar get_va_name(SqlNumber p_order_status_code);

		abstract void colour_row_by_status(SqlVarchar p_order_status_text_item_name, SqlNumber p_order_status_code);
	}

	abstract class Frm {

		abstract void fetch_product_statuses_list();

		abstract void populate_form_data();

		abstract void update_profit();

		abstract void populate_orders_data();

		abstract void populate_warehouses_data();

		abstract void new_form_instance();
	}

	@Package
	class FrmBody extends Frm {

		void fetch_product_statuses_list() {
			SqlVarchar v_list_item_name = "PRODUCT_INFORMATION.PRODUCT_STATUS";
			RecordGroup v_product_statuses_rg = RecordGroup.NULL();
			Item v_product_statuses_list_item = findItem(v_list_item_name);
			SqlNumber v_result;

			v_product_statuses_rg = findGroup("PRODUCT_STATUSES_RG");
			clearList(v_product_statuses_list_item);
			v_result = populateGroup(v_product_statuses_rg);
			populateList(v_list_item_name, v_product_statuses_rg);
		}

		void populate_form_data() {
			SqlVarchar v_custom_supplier_name = "Tech Comp Electronics Inc.";
			SqlVarchar v_custom_quality_class = "High";

			copy(v_custom_supplier_name, "PRODUCT_INFORMATION.SUPPLIER_NAME");
			copy(v_custom_quality_class, "PRODUCT_INFORMATION.QUALITY_CLASS");
			update_profit();
		}

		void update_profit() {
			SqlNumber v_profit = 0;
			SqlNumber v_list_price = numberIn("PRODUCT_INFORMATION.LIST_PRICE");
			SqlNumber v_min_price = numberIn("PRODUCT_INFORMATION.MIN_PRICE");

			v_profit = v_list_price - v_min_price;
			copy(v_profit, "PRODUCT_INFORMATION.PROFIT_VALUE");
			if (v_profit > 0) {
				setItemProperty("PRODUCT_INFORMATION.PROFIT_VALUE", VISUAL_ATTRIBUTE, "PROFIT_POSITIVE");
			} else {
				setItemProperty("PRODUCT_INFORMATION.PROFIT_VALUE", VISUAL_ATTRIBUTE, "PROFIT_NEGATIVE");
			}
			setItemProperty("PRODUCT_INFORMATION.PROFIT_VALUE", PROMPT_VISUAL_ATTRIBUTE, "PROFIT_PROMPT");
		}

		void populate_orders_data_customer() {
			sqlSelectInto(nameIn("ORDER_ITEMS.ORDER_CUSTOMER_NAME"),
				"SELECT c.cust_first_name || ' ' || c.cust_last_name \n" + 
				"FROM orders o, customers c \n" +
				"WHERE o.customer_id = c.customer_id \n" +
				"AND o.order_id = :ORDER_ITEMS.order_id"
			);
		}

		void populate_orders_data_status() {
			sqlSelectInto(nameIn("ORDER_ITEMS.ORDER_STATUS"), 
				"SELECT o.order_status \n" + 
				"FROM orders o \n" +
				"WHERE o.order_id = :ORDER_ITEMS.order_id"
			);

			copy(order_status.get_order_status_name(numberIn("ORDER_ITEMS.ORDER_STATUS")), "ORDER_ITEMS.ORDER_STATUS_NAME");
			order_status.colour_row_by_status("ORDER_STATUS_NAME", numberIn("ORDER_ITEMS.ORDER_STATUS"));
		}

		void populate_orders_data_date() {
			sqlSelectInto(nameIn("ORDER_ITEMS.ORDER_DATE"), 
				"SELECT TO_CHAR(o.order_date, 'YYYY-MM-DD HH24:MI:SS') \n" +
				"FROM orders o \n" + 
				"WHERE o.order_id = :ORDER_ITEMS.order_id"
			);
		}

		void populate_orders_data() {
			populate_orders_data_customer();
			populate_orders_data_status();
			populate_orders_data_date();
		}

		void populate_warehouses_data() {
			sqlSelectInto(nameIn("INVENTORIES.WAREHOUSE_NAME"), 
				"SELECT warehouse_name \n" +
				"FROM warehouses \n" + 
				"WHERE warehouse_id = :INVENTORIES.WAREHOUSE_ID"
			);
		}

		void new_form_instance() {
			if (isNotNull(nameIn("PARAMETER.PRODUCT_ID"))) {
				setBlockProperty("PRODUCT_INFORMATION", DEFAULT_WHERE, "PRODUCT_ID = :PARAMETER.PRODUCT_ID");
			}
			fetch_product_statuses_list();
			goBlock("PRODUCT_INFORMATION");
			executeQuery();
			populate_form_data();
		}
	}

	@Package
	class UtilsBody extends Utils {

		ParameterList get_empty_param_list() {
			ParameterList v_input_params = ParameterList.NULL();

			v_input_params = getParameterList("input_params");
			if (!isNull(v_input_params)) {
				destroyParameterList(v_input_params);
			}
			v_input_params = createParameterList("input_params");
			return v_input_params;
		}

		ParameterList add_param_to_list(ParameterList p_input_params, SqlVarchar p_name, SqlNumber p_value) {
			addParameter(p_input_params, p_name, TEXT_PARAMETER, p_value);
			return p_input_params;
		}
	}

	abstract class Utils {

		abstract ParameterList get_empty_param_list();

		abstract ParameterList add_param_to_list(ParameterList p_input_params, SqlVarchar p_name, SqlNumber p_value);
	}

	@Package
	class AppRunnerBody extends AppRunner {

		void run_app(SqlVarchar p_app_name, ParameterList p_input_params) {
			SqlVarchar v_form_name;

			callForm(p_app_name, NO_HIDE, NO_REPLACE, NO_QUERY_ONLY, p_input_params);
		}

		void run_app(SqlVarchar p_app_name) {
			ParameterList p_input_params = ParameterList.NULL();

			p_input_params = utils.get_empty_param_list();
			run_app(p_app_name, p_input_params);
		}
	}

	abstract class AppRunner {

		abstract void run_app(SqlVarchar p_app_name);

		abstract void run_app(SqlVarchar p_app_name, ParameterList p_input_params);
	}

	@Package
	class FormOperationsBody extends FormOperations {

		void commit_current_form() {
			SqlNumber v_current_msg_level;

			// pre commit --
			v_current_msg_level = numberIn("SYSTEM.MESSAGE_LEVEL");
			copy(15, "SYSTEM.MESSAGE_LEVEL");
			doKey("ENTER");
			if (!formSuccess()) {
				throw new FormTriggerFailureException();
			}

			// Commit if anything is changed
			if (nameIn("SYSTEM.FORM_STATUS") == "CHANGED") {
				doKey("COMMIT_FORM");
				if (nameIn("SYSTEM.FORM_STATUS") != "QUERY") {
					message("An error prevented your changes from being committed.");
					throw new FormTriggerFailureException();
				}
			}

			// post commit --
			copy(v_current_msg_level, "SYSTEM.MESSAGE_LEVEL");
		}

		void exit_current_form() {
			clearForm(NO_VALIDATE);
			doKey("EXIT_FORM");
		}
	}

	abstract class FormOperations {

		abstract void commit_current_form();

		abstract void exit_current_form();
	}

	void check_package_failure() {
		if (! formSuccess()) {
			throw new FormTriggerFailureException();
		}
	}

	void query_master_details(Relation rel_id, SqlVarchar detail) {
		SqlVarchar oldmsg; // Old Message Level Setting
		SqlVarchar reldef; // Relation Deferred Setting

		try {
			//
			// Initialize Local Variable(s)
			//
			reldef = getRelationProperty(rel_id, DEFERRED_COORDINATION);
			oldmsg = nameIn("System.Message_Level");
			//
			// If NOT Deferred, Goto detail and execute the query.
			//
			if (reldef == "FALSE") {
				goBlock(detail);
				check_package_failure();
				copy("10", "System.Message_Level");
				executeQuery();
				copy(oldmsg, "System.Message_Level");
			} else {
				//
				// Relation is deferred, mark the detail block as un-coordinated
				//
				setBlockProperty(detail, COORDINATION_STATUS, NON_COORDINATED);
			}

		} catch (FormTriggerFailureException e1) {
			copy(oldmsg, "System.Message_Level");
			throw e1;
		}
	}

	void clear_all_master_details() {
		class sub$clear_all_master_details {
			SqlVarchar mastblk; // Initial Master Block Causing Coord
			SqlVarchar coordop; // Operation Causing the Coord
			SqlVarchar trigblk; // Cur Block On-Clear-Details Fires On
			SqlVarchar startitm; // Item in which cursor started
			SqlVarchar frmstat; // Form Status
			SqlVarchar curblk; // Current Block
			SqlVarchar currel; // Current Relation
			SqlVarchar curdtl; // Current Detail Block

			private SqlVarchar first_changed_block_below(SqlVarchar master) {
				SqlVarchar curblk; // Current Block
				SqlVarchar currel; // Current Relation
				SqlVarchar retblk; // Return Block

				//
				// Initialize Local Vars
				//
				curblk = master;
				currel = getBlockProperty(curblk, FIRST_MASTER_RELATION);
				//
				// While there exists another relation for this block
				//
				while (isNotNull(currel)) {
					//
					// Get the name of the detail block
					//
					curblk = getRelationProperty(currel, DETAIL_NAME);
					//
					// If this block has changes, return its name
					//
					if (getBlockProperty(curblk, STATUS) == "CHANGED") {
						return curblk;
					} else {
						//
						// No changes, recursively look for changed blocks below
						//
						retblk = first_changed_block_below(curblk);
						//
						// If some block below is changed, return its name
						//
						if (isNotNull(retblk)) {
							return retblk;
						} else {
							//
							// Consider the next relation
							//
							currel = getRelationProperty(currel, NEXT_MASTER_RELATION);
						}
					}
				}

				//
				// No changed blocks were found
				//
				return SqlVarchar.NULL();
			}

			void call() {
				try {
					//
					// Init Local Vars
					//
					mastblk = nameIn("System.Master_Block");
					coordop = nameIn("System.Coordination_Operation");
					trigblk = nameIn("System.Trigger_Block");
					startitm = nameIn("System.Cursor_Item");
					frmstat = nameIn("System.Form_Status");

					//
					// If the coordination operation is anything but CLEAR_RECORD or
					// SYNCHRONIZE_BLOCKS, then continue checking.
					//
					if (coordop.notIn("CLEAR_RECORD", "SYNCHRONIZE_BLOCKS")) {
						//
						// If we're processing the driving master block...
						//
						if (mastblk == trigblk) {
							//
							// If something in the form is changed, find the
							// first changed block below the master
							//
							if (frmstat == "CHANGED") {
								curblk = first_changed_block_below(mastblk);
								//
								// If we find a changed block below, go there
								// and Ask to commit the changes.
								//
								if (isNotNull(curblk)) {
									goBlock(curblk);
									check_package_failure();
									clearBlock(ASK_COMMIT);
									//
									// If user cancels commit dialog, raise error
									//
									if (!(nameIn("System.Form_Status") == "QUERY" || nameIn("System.Block_Status") == "NEW")) {
										throw new FormTriggerFailureException();
									}
								}
							}
						}
					}

					//
					// Clear all the detail blocks for this master without
					// any further asking to commit.
					//
					currel = getBlockProperty(trigblk, FIRST_MASTER_RELATION);
					while (isNotNull(currel)) {
						curdtl = getRelationProperty(currel, DETAIL_NAME);
						if (getBlockProperty(curdtl, STATUS) != "NEW") {
							goBlock(curdtl);
							check_package_failure();
							clearBlock(NO_VALIDATE);
							if (nameIn("System.Block_Status") != "NEW") {
								throw new FormTriggerFailureException();
							}
						}
						currel = getRelationProperty(currel, NEXT_MASTER_RELATION);
					}

					//
					// Put cursor back where it started
					//
					if (nameIn("System.Cursor_Item") != startitm) {
						goItem(startitm);
						check_package_failure();
					}

				} catch (FormTriggerFailureException e1) {
					if (nameIn("System.Cursor_Item") != startitm) {
						goItem(startitm);
					}
					throw e1;
				}
			}
		}

		new sub$clear_all_master_details().call();
	}
}
